;;
;;this is a text file
;;

;; Test expressions
;;
(define e1 '((a and b) or (c and d)))


(define e2 '(((a or b) and c) or (d and e)))


(define e3 '(((a and b) or (c and d)) or ((e and f) and (g or h))))


(define e4 '(not (a or b)))


(define e5 (list 'not e2))


(define e6 (list 'not e3))


;;
;;  Function cnf! (Conjunctive Normal Form)
;;
;;  I 	      	cnf[ p ] = p
;;  II	      	cnf[ (~ p) ] = neg[ p ]
;;  III      	cnf[ (^ P Q) ] = (^ cnf[ P ] cnf[ Q ])
;;  IV       	cnf[ (v P (^ Q R)) ] = cnf[ (^ (v P Q) (v P R)) ]
;;  V      	cnf[ (v (^ Q R) P) ] = cnf[ (v P (^ Q R)) ]
;;  VI   	cnf[ (v P Q) ] = (v cnf[P] cnf[Q]) 		iff P=cnf[P] and Q=cnf[Q]
;;                                       = cnf[ (v cnf[P] cnf[Q]) ]  	otherwise
;;  VII     	cnf[ (~ (v P Q)) ]  = (^ cnf[ (~P) ] cnf[ (~Q) ] )  
;;  VIII   	cnf[ (~ (^ P Q)) ] = cnf[ (v cnf[ (~P) ] cnf[ (~Q) ]) ]
;;  IX		cnf[ (~ P) ]  = (~ cnf[ P ])			iff P=cnf[P]
;;                                    = cnf[ (~ cnf[P]) ]		otherwise
;;
(define cnf! 
   (lambda (e)
       (infix  (cnf (prefix e)))))


(define cnf 
   (lambda (e)	     
        (cond ((rule1? e) (do-rule1 e))
                    ((rule2? e) (do-rule2 e))
                    ((rule3? e) (do-rule3 e))
                    ((rule4? e) (do-rule4 e))
                    ((rule5? e) (do-rule5 e))
                    ((rule6? e) (do-rule6 e))
                    ((rule7? e) (do-rule7 e))
                    ((rule8? e) (do-rule8 e))
                    ((rule9? e) (do-rule9 e))
                    (else 'error))))


;; Rule 1: cnf[ p ] = p
;;
(define rule1? 
    (lambda (e) (symbol? e)))


(define do-rule1 
    (lambda (e) e))


;; Rule 2: cnf[ (~p) ] = neg[p]  
;;
(define rule2? 
    (lambda (e) (and (list? e) (eq? (car e) 'not) (symbol? (cadr e)))))


(define do-rule2
    (lambda (e) (neg e)))


;; Rule 3: cnf[ (^ P Q) ] = (^ cnf[P] cnf[Q]) 
;;
(define rule3? 
    (lambda (e) (and (eq? (car e) 'and))))


(define do-rule3 
    (lambda (e)
        (let ( (p (cadr e)) (q (caddr e)) )
                (list 'and (cnf  p) (cnf q)))))


;; Rule 4: cnf[ (v P (^ Q R)) ]  = cnf[ (^ (v P Q) (v P R)) ]
;;
(define rule4? 
    (lambda (e)
        (and (eq? (car e) 'or) (list? (caddr e)) (eq? (car (caddr e)) 'and))))


(define do-rule4 
    (lambda (e)
       (let ( (p (cadr e)) (q (cadr (caddr e))) (r (cadr (cdr (caddr e)))) )
               (list 'and
                  (cnf (list 'or p q))
                  (cnf (list 'or p r))))))


;; Rule 5: cnf[ (v (^ Q R) P) ] = cnf[ ( v P (^ Q R)) ]
;;
(define rule5? 
    (lambda (e)
        (and (eq? (car e) 'or) (list? (cadr e)) (eq? (car (cadr e)) 'and))))


(define do-rule5 
    (lambda (e)
       (let ( (p (caddr e)) (qr (cadr e)) )
               (cnf (list 'or p qr)))))


;; Rule 6: cnf[ (v P Q) ] = (v cnf[P] cnf[Q])   iff cnf[P]=P and cnf[Q]=Q
;;                                           cnf[ (v cnf[P] cnf[Q]) ]  otherwise
;;
(define rule6? 
     (lambda (e) 
           (and (eq? (car e) 'or)))) 


(define do-rule6 
     (lambda (e)
         (let* ( (p (cadr e)) (q (caddr e)) (pp (cnf p)) (qq (cnf q)) )
               (if (and (equal? p pp) (equal? q qq))
                    (list 'or pp qq) 
                    (cnf (list 'or (cnf p) (cnf q)))))))


;; Rule 7: cnf[ (~ (v P Q)) ] = cnf[ (^ (~P) (~Q) ) ] 
;;
(define rule7? 
    (lambda (e)
        (and (eq? (car e) 'not) (list? (cadr e)) (eq? (car (cadr e)) 'or))))


(define do-rule7 
     (lambda (e)
        (let* ( (pq (cadr e)) (p (cadr pq)) (q (caddr pq)) )
                (cnf (list 'and (list 'not p) (list 'not q))))))


;; Rule 8: cnf[ (~ (^ P Q)) ] = cnf[ (v (~P) (~Q) ) ] 
;;
(define rule8? 
    (lambda (e)
        (and (eq? (car e) 'not) (list? (cadr e)) (eq? (car (cadr e)) 'and))))


(define do-rule8 
     (lambda (e)
        (let* ( (pq (cadr e)) (p (cadr pq)) (q (caddr pq)) )
                 (cnf (list 'or (list 'not p) (list 'not q))))))


;; Rule 9: cnf[ (~P) ] = (~ cnf[P]) iff P=cnf[P]
;;                                   = cnf[ (~ cnf[P]) ]	otherwise
;;
(define rule9? 
   (lambda (e)
        (and (eq? (car e) 'not) (list? (cadr e)))))


(define do-rule9 
    (lambda (e)
        (let* ( (p (cadr e)) (pp (cnf p)) )
             (if (equal? p pp)
                  (list 'not pp)
                  (cnf (list 'not pp)))))) 


;; Prefix notation
;;
(define prefix 
    (lambda (e)
        (cond ((symbol? e) e)
                    ((eq? (car e) 'not) (list 'not (prefix (cadr e))))
                    (else (list (cadr e)  (prefix (car e)) (prefix (caddr e)))))))


;; Infix notation
;;
(define infix 
     (lambda (e)
           (cond ((symbol? e) e)
                       ((eq? (car e) 'not) (list 'not (infix (cadr e))))
                       (else (list (infix (cadr e)) (car e) (infix (caddr e)))))))                   


;; Temporary
(define neg
     (lambda (e) e))


;; Trace
;; (trace do-rule1 do-rule2 do-rule3 do-rule4 do-rule5)
;; (trace do-rule6 do-rule7 do-rule8 do-rule9)




